# Generated by Django 2.0 on 2018-01-21 18:25

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='IntegerSequencedRecord',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('sequence_id', models.UUIDField()),
                ('position', models.BigIntegerField()),
                ('topic', models.TextField()),
                ('data', models.TextField()),
            ],
            options={
                'db_table': 'integer_sequenced_items',
            },
        ),
        migrations.CreateModel(
            name='SnapshotRecord',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('sequence_id', models.UUIDField()),
                ('position', models.BigIntegerField()),
                ('topic', models.TextField()),
                ('data', models.TextField()),
            ],
            options={
                'db_table': 'snapshots',
            },
        ),
        migrations.CreateModel(
            name='StoredEventRecord',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('originator_id', models.UUIDField()),
                ('originator_version', models.BigIntegerField()),
                ('event_type', models.TextField()),
                ('state', models.TextField()),
            ],
            options={
                'db_table': 'stored_events',
            },
        ),
        migrations.CreateModel(
            name='TimestampSequencedRecord',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('sequence_id', models.UUIDField()),
                ('position', models.DecimalField(decimal_places=6, max_digits=24)),
                ('topic', models.TextField()),
                ('data', models.TextField()),
            ],
            options={
                'db_table': 'timestamp_sequenced_items',
            },
        ),
        migrations.AddIndex(
            model_name='timestampsequencedrecord',
            index=models.Index(fields=['position'], name='position_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='timestampsequencedrecord',
            unique_together={('sequence_id', 'position')},
        ),
        migrations.AlterUniqueTogether(
            name='storedeventrecord',
            unique_together={('originator_id', 'originator_version')},
        ),
        migrations.AlterUniqueTogether(
            name='snapshotrecord',
            unique_together={('sequence_id', 'position')},
        ),
        migrations.AlterUniqueTogether(
            name='integersequencedrecord',
            unique_together={('sequence_id', 'position')},
        ),
    ]
